/* HTTP server
 * A nearly complete rewrite, extensively expanded, based loosely on
 * the HTTP server of Selcuk Ozturk <seost2+@pitt.edu> &
 * <ashok@biochemistry.cwru.edu>, which were based on
 * Chris McNeil's <cmcneil@mta.ca> Gopher server, which was based in
 * part on fingerd.c
 *
 * If using MSDOS, the 'path' can be /drive:/path
 */

#include "global.h"
#include "ctype.h"
#include "commands.h"
#include <time.h>
#ifdef MSDOS
#include <dir.h>
#include <dos.h>
#include "hardware.h"
#endif
#ifdef HTTP
#include <sys/stat.h>
#include "files.h"
#include "mbuf.h"
#include "socket.h"
#include "session.h"
#include "mailbox.h"
#include "smtp.h"


#if !defined(_lint) && !defined(MSDOS)
static char rcsid[] OPTIONAL = "$Id: http.c,v 1.16 1996/07/24 00:54:28 root Exp root $";
#endif

#define DEFAULT_ERROR_MSG "[an error occurred while processing this directive]"
#define HTTP_TIME_FORMAT "%a, %d %b %Y %T GMT\n"
#define DEFAULT_TIME_FORMAT "%A, %d-%b-%y %T %Z"
#define SIZEFMT_BYTES 0
#define SIZEFMT_KMG 1

static char entry[] = "<IMG SRC=\"/icons/%s.xbm\" ALT=\"%s\"> ";
#ifdef MSDOS
static char entry1[] = "<A HREF=\"/%c:%s%s\">%-s</a>";
static char DR = 'c';
#else
static char entry1[] = "<A HREF=\"/%s%s\">%-s</a>";
#endif
static char entry2[] = "%2d-%s-%02d %02d:%02d %6dK\n";
int HttpUsers = 0;
static int Httpmax = 10;
static int Httpsimc = 5;
static int Httptdisc = 180;
static int HTTPreferer = 1;
static int HTTPagent = 1;
static int HTTPmisc = 1;
static int HTTPpost = 1;
#ifdef UNIX
static int HTTPexeccmd = 0;
#endif
#ifdef HTTPPBBS
static int HTTPexecbbs = 0;
static int HTTPanonsend = 0;
#endif
#if 0
static int HTTPscript = 0;
#endif
static FILE * HTTPsavedfp;


struct reqInfo {
#if 0
	int		index;
	int		response;
#endif
	int		qsize;		/* not used at the present */
	int		sizefmt;
	int		isvolatile;
#ifdef HTTPPBBS
	int		msgcount;
	int		msgnum;
	int		anony;
	long		privs;
#endif
	char		*url;
	char		*method;
	char		*version;
	char		*newcheck;
	char		*from;
	char		*referer;
	char		*agent;
	char		*passwd;
	char		*dirname;
	char 		*usetime;
	char		*useerror;
	char		buf[PLINELEN + 16];
#ifdef HTTPPBBS
	char		*areaname;
	char		*username;
	struct mbx 	*m;
	char		*bbspath;
#endif
};


/* function prototypes */
static int dohttpagent __ARGS((int argc,char *argv[],void *p));
#ifdef MSDOS
static int dohttpdrive __ARGS((int argc,char *argv[],void *p));
#endif
#ifdef UNIX
static int dohttpexeccmd __ARGS((int argc,char *argv[],void *p));
#endif
#ifdef HTTPPBBS
static int dohttpexecbbs __ARGS((int argc,char *argv[],void *p));
static int checkareaperms __ARGS((struct reqInfo *rq));
#endif
static int dohttpmax __ARGS((int argc,char *argv[],void *p));
static int dohttpmisc __ARGS((int argc,char *argv[],void *p));
static int dohttppost __ARGS((int argc,char *argv[],void *p));
static int dohttpreferer __ARGS((int argc,char *argv[],void *p));
#if 0
static int dohttpscript __ARGS((int argc,char *argv[],void *p));
#endif
static int dohttpsim __ARGS((int argc,char *argv[],void *p));
static int dohttpstatus __ARGS((int argc,char *argv[],void *p));
static int dohttptdisc __ARGS((int argc,char *argv[],void *p));

static void httpserv __ARGS((int s,void *unused,void *p));
static int mkwelcome __ARGS((const char *Hdir,char *file));
static char * decode __ARGS((char *string));
static void httpHeaders __ARGS((int type, const char *str, struct reqInfo *rq, int more, const char *realm));
static void httpFileinfo __ARGS((FILE *fp, struct reqInfo *rq, int thetype, char *vers));
static void openHTTPlog __ARGS((const char *name));
static void closeHTTPlog __ARGS((void));
static int getmonth __ARGS((char *cp));
static int isnewer __ARGS((time_t thetime,char *tmstr));
static long countusage __ARGS((const char *file, const char *basedir, int display, int increase));
static void sendhtml __ARGS((FILE *fp, struct reqInfo *rq));
static void file_or_virtual __ARGS((char *cp, char *newname, struct reqInfo *rq));
static void send_size __ARGS((int sizefmt,long size));
static char *ht_time __ARGS((time_t t,const char *fmt,int gmt));
char *gmt_ptime __ARGS((time_t *thetime));
static int addFile __ARGS((const char *fname, int s));
#ifdef HTTPPBBS
static void addUserFile __ARGS((struct reqInfo *rq));
static void addMsgFile __ARGS((char *file, struct reqInfo *rq));
static void doPBBScmd __ARGS((struct reqInfo *rq, const char *cmdstring));
static void doPBBScmd_init __ARGS((struct reqInfo *rq));
static void doPBBScmd_term __ARGS((struct reqInfo *rq));
static int dohttpanonsend __ARGS((int argc,char *argv[],void *p));
#endif

extern char *base64ToStr __ARGS((char *b64));
#ifdef ALLSERV
extern char *getquote __ARGS((void));
#endif

/* HTTP Header lines sent or processed */
#define HDR_TYPE		0
#define HDR_LENGTH		1
#define HDR_MIME		2
#define HDR_SERVER		3
#define HDR_LOCATION		4
#define HDR_DATE		5
#define HDR_MODIFIED		6
#define HDR_SINCE		7
#define HDR_REFERER		8
#define HDR_AGENT		9
#define HDR_FROM		10
#define HDR_EXPIRES		11
#define HDR_AUTHORIZE		12
#define HDR_AUTHENTICATE	13

static const char *HTTPHdrs[] = {
	"Content-Type:",
	"Content-Length:",
	"MIME-version:",
	"Server:",
	"Location:",
	"Date:",
	"Last-Modified:",
	"If-Modified-Since:",
	"Referer:",
	"User-Agent:",
	"From:",
	"Expires:",
	"Authorization:",
	"WWW-Authenticate:"
};


/* First component of MIME types */

static const char *CTypes1[] = {
	"video/",
	"text/",
	"image/",
	"audio/",
	"application/"
};


/* Struct for looking up MIME types from file extensions */
struct FileTypes {
	const char *ext;
	short type1;
	const char *type2;
};


static struct FileTypes DFAR HTTPtypes[] = {
	/* Other 'special' entries can be added in front of 'html',
	 * but care must be taken to make sure that the defines for
	 * HTTPbinary, HTTPplain, HTTPform, and HTTPhtml are correct
	 * and that no entries with a 'NULLCHAR' ext come after 'html'
	 * except for the one that ends the table.
	 */
#define HTTPbinary 0
	{ NULLCHAR,	4,	"octet-stream" },
#define HTTPplain 1
	{ NULLCHAR,	1,	"plain" },
#define HTTPform 2
	{ NULLCHAR,	4,	"x-www-form-urlencoded" },
#define HTTPhtml 3
	{ "html",	1,	"html" },
	/* the following can be in any order */
	{ "htm",	1,	"html" },
	{ "avi",	0,	"x-msvideo" },
	{ "qt",		0,	"quicktime" },
	{ "mov",	0,	"quicktime" },
	{ "mpeg",	0,	"mpeg" },
	{ "mpg",	0,	"mpeg" },
	{ "rtx",	1,	"richtext" },
	{ "xpm",	2,	"x-xpixmap" },
	{ "xbm",	2,	"x-xbitmap" },
	{ "rgb",	2,	"x-rgb" },
	{ "ppm",	2,	"x-portable-pixmap" },
	{ "pgm",	2,	"x-portable-graymap" },
	{ "pbm",	2,	"x-portable-bitmap" },
	{ "pnm",	2,	"x-portable-anymap" },
	{ "ras",	2,	"x-cmu-raster" },
	{ "tiff",	2,	"tiff" },
	{ "tif",	2,	"tiff" },
	{ "jpeg",	2,	"jpeg" },
	{ "jpg",	2,	"jpeg" },
	{ "gif",	2,	"gif" },
	{ "wav",	3,	"x-wav" },
	{ "aiff",	3,	"x-aiff" },
	{ "aif",	3,	"x-aiff" },
	{ "au",		3,	"basic" },
	{ "snd",	3,	"basic" },
	{ "tar",	4,	"x-tar" },
	{ "man",	4,	"x-trof-man" },
	{ "rtf",	4,	"rtf" },
	{ "eps",	4,	"postscript" },
	{ "ps",		4,	"postscript" },
	{ "sit",	4,	"x-stuffit" },
	{ "hqx",	4,	"mac-binhex40" },
	{ "fif",	4,	"fractals" },
	{ "zip",	4,	"x-zip" },
	{ "gz",		4,	"x-gzip" },
	{ "z",		4,	"x-compress" },
	{ NULLCHAR,	-1,	NULLCHAR }
};


/* Struct to keep tract of ports defined and their root directory */

#define MAXHTTPPORTS 10
struct portsused {
	int	port;
	int	sock;
	char	*dirname;
};
static struct portsused ports[MAXHTTPPORTS];


#ifdef UNIX
#define OS "Unix"
#define WILDCARD "*"
#else
#define OS "MS-DOS"
#define WILDCARD "*.*"
#endif



static struct cmds DFAR Httpcmds[] = {
	{ "agent",	dohttpagent,	0, 0, NULLCHAR },
#ifdef HTTPPBBS
	{ "anonsend",	dohttpanonsend,	0, 0, NULLCHAR },
#endif
#ifdef MSDOS
	{ "drive",	dohttpdrive,	0, 0, NULLCHAR },
#endif
#ifdef HTTPPBBS
	{ "execbbs",	dohttpexecbbs,	0, 0, NULLCHAR },
#endif
#ifdef UNIX
	{ "execcmd",	dohttpexeccmd,	0, 0, NULLCHAR },
#endif
	{ "maxcli",	dohttpmax,	0, 0, NULLCHAR },
	{ "misc",	dohttpmisc,	0, 0, NULLCHAR },
	{ "post",	dohttppost,	0, 0, NULLCHAR },
	{ "referer",	dohttpreferer,	0, 0, NULLCHAR },
#if 0
	{ "script",	dohttpscript,	0, 0, NULLCHAR },
#endif
	{ "simult",	dohttpsim,	0, 0, NULLCHAR },
	{ "status",	dohttpstatus,	0, 0, NULLCHAR },
	{ "tdisc",	dohttptdisc,	0, 0, NULLCHAR },
	{ NULLCHAR,	NULL,		0, 0, NULLCHAR }
};


int
dohttp (argc, argv, p)
int argc;
char *argv[];
void *p;
{
	return subcmd (Httpcmds, argc, argv, p);
}


int
dohttpstatus (argc, argv, p)
int argc OPTIONAL;
char *argv[];
void *p;
{
int found = 0, k;

	for (k = 0; k < MAXHTTPPORTS; k++)	{
		if (ports[k].port)	{
			found++;
			tprintf ("HTTP Server active on port #%-4d - %s\n", ports[k].port, ports[k].dirname);
		}
	}
	if (!found)
		tputs ("No");
	else
		tprintf ("%d", found);
	tprintf (" HTTP Server%s %s active\n\n", (found == 1) ? "" : "s", (found == 1) ? "is" : "are");

	(void) dohttpagent (0, argv, p);
#ifdef MSDOS
	(void) dohttpdrive (0, argv, p);
#endif
	(void) dohttpmax (0, argv, p);
	(void) dohttpmisc (0, argv, p);
	(void) dohttppost (0, argv, p);
	(void) dohttpreferer (0, argv, p);
	(void) dohttpsim (0, argv, p);
	(void) dohttptdisc (0, argv, p);
	return 0;
}


#ifdef MSDOS
static int
dohttpdrive (argc, argv, p)
int argc;
char *argv[];
void *p OPTIONAL;
{
	if (argc < 2)
		tprintf ("Def. drive: %c\n",DR);
	else if (!isalpha (*argv[1]))
		tputs ("Bad drive letter\n");
	else
		DR = *argv[1];
	return 0;
}
#endif


#ifdef UNIX
static int
dohttpexeccmd (argc, argv, p)
int argc;
char *argv[];
void *p OPTIONAL;
{
	return setbool (&HTTPexeccmd, "Enable Server-side include 'exec cmd's", argc, argv);
}
#endif


#ifdef HTTPPBBS
static int
dohttpexecbbs (argc, argv, p)
int argc;
char *argv[];
void *p OPTIONAL;
{
	return setbool (&HTTPexecbbs, "Enable Server-side include 'exec bbs's", argc, argv);
}


static int
dohttpanonsend (argc, argv, p)
int argc;
char *argv[];
void *p OPTIONAL;
{
	/* related to the NO_HTTP_MAIL permission */
	return setbool (&HTTPanonsend, "Allow anonymous users to SEND mail", argc, argv);
}
#endif


#if 0
static int
dohttpscript (argc, argv, p)
int argc;
char *argv[];
void *p OPTIONAL;
{
	return setbool (&HTTPscript, "Enable Server-side include 'tscript's", argc, argv);
}
#endif


static int
dohttpreferer (argc, argv, p)
int argc;
char *argv[];
void *p OPTIONAL;
{
	return setbool (&HTTPreferer, "Log Referer Headers", argc, argv);
}


static int
dohttpagent (argc, argv, p)
int argc;
char *argv[];
void *p OPTIONAL;
{
	return setbool (&HTTPagent, "Log User-Agent Headers", argc, argv);
}


static int
dohttpmisc (argc, argv, p)
int argc;
char *argv[];
void *p OPTIONAL;
{
	return setbool (&HTTPmisc, "Log Misc Headers", argc, argv);
}


static int
dohttppost (argc, argv, p)
int argc;
char *argv[];
void *p OPTIONAL;
{
	return setbool (&HTTPpost, "Log POST Requests", argc, argv);
}


static int
dohttpmax (argc, argv, p)
int argc;
char *argv[];
void *p OPTIONAL;
{
	return setint (&Httpmax, "Max. HTTP connects", argc, argv);
}


static int
dohttpsim (argc, argv, p)
int argc;
char *argv[];
void *p OPTIONAL;
{
	return setint (&Httpsimc, "Simult. HTTP conn.'s serviced", argc, argv);
}


static int
dohttptdisc (argc, argv, p)
int argc;
char *argv[];
void *p OPTIONAL;
{
	return setint (&Httptdisc, "HTTP Server tdisc (sec)", argc, argv);
}


/* Start up http service */
int
httpstart (argc, argv, p)
int argc;
char *argv[];
void *p OPTIONAL;
{
int port, k;

	if(argc < 2)
		port = IPPORT_HTTP;
	else
		port = atoi(argv[1]);

	for (k = 0; k < MAXHTTPPORTS; k++) {
		if (ports[k].port == port)	{
			tprintf ("Sorry, but you already have an HTTP server on port #%d!\n", port);
			return 0;
		}
	}
	for (k = 0; k < MAXHTTPPORTS; k++) {
		if (ports[k].sock <= 0)
			break;
	}
	if (k == MAXHTTPPORTS)	{
		tprintf ("Sorry, but all %d HTTP ports are assigned!\n", MAXHTTPPORTS);
		return 0;
	}

	if(argc < 3)
		ports[k].dirname = strdup (HTTPdir);
	else
		ports[k].dirname = strdup (argv[2]);
	ports[k].sock = -1;
	ports[k].port = port;
	return (installserver (argc, argv, &ports[k].sock, "HTTP Listener", IPPORT_HTTP,
		"HTTP Server", httpserv, 1024, (void *)k));
}


int
http0 (argc, argv, p)
int argc OPTIONAL;
char *argv[] OPTIONAL;
void *p OPTIONAL;
{
int port, k;

	if(argc < 2)
		port = IPPORT_HTTP;
	else
		port = atoi(argv[1]);
	for (k = 0; k < MAXHTTPPORTS; k++) {
		if (ports[k].port == port)	{
			ports[k].port = 0;
			free (ports[k].dirname);
			ports[k].dirname = NULLCHAR;
			return (deleteserver (&ports[k].sock));
		}
	}
	tprintf ("Sorry, but no HTTP server was found on port #%d!\n", port);
	return 0;
}


/* Creates a copy of string which has '%' escaped charecters decoded
   back to ASCII. Allocates memory. Caller must free it. */


static char *
decode (string)
char *string;
{
int i,k,slen;
char code[3],*tmp;

	slen = strlen (string);
	tmp = mallocw (slen + 1);
	i = k = 0;

	while (i <= slen) {
		if (string[i] == '%') {
			string[i] = '#';        /* '%' sign causes problems in log()  */
			code[0] = string[i+1];
			code[1] = string[i+2];
			code[2] = '\0';
			tmp[k] = (char)htoi(code);
			k++;  i += 3;
		} else
			tmp[k++] = string[i++];
	}
	return tmp;
}


#ifdef HTTPPBBS
static void
doPBBScmd (rq, cmdstr)
struct reqInfo *rq;
const char *cmdstr;
{
	doPBBScmd_init (rq);
	strcpy (rq->m->line, cmdstr);
	doPBBScmd_term (rq);
}


static void
doPBBScmd_init (rq)
struct reqInfo *rq;
{
	rq->m = newmbx (1);
	rq->m->privs = rq->privs;
	rq->m->sid = MBX_HTTP;
	rq->m->user = Curproc->output;
	rq->m->path = rq->bbspath;
	if (!rq->username)
		strcpy (rq->username, "anonymous");
	strncpy (rq->m->name, rq->username, 19);
	if (rq->areaname)
		changearea (rq->m, rq->areaname, 0);
}

static void
doPBBScmd_term (rq)
struct reqInfo *rq;
{
	if (rq->m)	{
		(void) mbx_parse (rq->m);
		rq->m->path = NULLCHAR;
		exitbbs(rq->m);
		rq->m = NULLMBX;
	}
}


static void
addUserFile (rq)
struct reqInfo *rq;
{
char fname[256];

	sprintf (fname, "%s/bbs/user.html", rq->dirname);
	addMsgFile (fname, rq);
}


static void
addMsgFile (fname, rq)
char *fname;
struct reqInfo *rq;
{
FILE *fp;

	if ((fp = fopen (fname, READ_BINARY)) != NULLFILE) {
		sendhtml (fp, rq);
		fclose (fp);
	}
}



static int
checkareaperms (rq)
struct reqInfo *rq;
{
int retval = 0;

	/* only return a 1 if (1) it's our area, (2) we are a sysop, or
	   (3) it is a public area. Else, return a zero
	 */

	if (!stricmp(rq->areaname, rq->username) || (rq->privs & SYSOP_CMD) || isarea(rq->areaname))
		retval = 1;
	return retval;
}

#endif


static int
addFile (fname, s)
const char *fname;
int s;
{
FILE *fp;

	if ((fp = fopen (fname, READ_BINARY)) != NULLFILE) {
		(void) sendfile (fp, s, IMAGE_TYPE, 0);
		fclose (fp);
		return 1;
	}
	return 0;
}


static void
httpserv (s, v1, p)
int s;
void *v1 OPTIONAL;
void *p OPTIONAL;
{
char command[PLINELEN], *cp, *newcommand, *tmp, *version;
FILE *fp;
int port, length, i, err = 1, head = 0, retval = 200;
struct stat sb;
int isdir = 0;
struct reqInfo *rq;
	
	(void) sockowner (s, Curproc);
	(void) sockmode (s, SOCK_ASCII);
	rq = (struct reqInfo *)calloc (1, sizeof (struct reqInfo));

	rq->sizefmt = SIZEFMT_KMG;
	rq->useerror = strdup (DEFAULT_ERROR_MSG);
	rq->usetime = strdup (DEFAULT_TIME_FORMAT);

	close_s(Curproc->output);
	close_s(Curproc->input);
	Curproc->output = Curproc->input = s;

	HttpUsers++;

	if (Httpmax < HttpUsers)	{
		tprintf("HTTP/1.0 503 Service Unavailable\n%s %s%s\n\n<TITLE>Busy</TITLE><H1>Busy</H1>\nTry again later.\n",HTTPHdrs[0],CTypes1[HTTPtypes[HTTPhtml].type1],HTTPtypes[HTTPhtml].type2);
		goto quit;
	}

	while (HttpUsers > Httpsimc)
		pwait(&HttpUsers);

	alarm (Httptdisc * 1000L);
	if (recvline (s, command, PLINELEN) == -1)
		goto quit;
	alarm (0L);

	port = ports[(int)v1].port;
	rq->dirname = ports[(int)v1].dirname;
	
	rip (command);
	if (!*command)	{
		retval = 400;
		usputs (s, "<HEAD><TITLE>400 Bad Request</TITLE></HEAD>\n<BODY><H1>400 Bad Request</H1>\nYour client sent a query that this server could not understand.<P>\nReason: Invalid HTTP/0.9 method.<P></BODY>\n");
		goto quit;
	}

	tmp = decode (command);
	cp = strchr (tmp, ' ');
	*cp++ = 0;
	newcommand = ++cp;
	version = strchr (newcommand, ' ');
	if (version != NULLCHAR)
		*version++ = 0;
	
	/* at this point, tmp=command, newcommand=ulr, and version=httpversion info */
	if (!strcmp (newcommand, "/"))
		newcommand++;
	rq->url = strdup (newcommand);
	rq->version = strdup (version);
	rq->method = strdup (tmp);
	length = strlen (newcommand);

	/* if HTTP 0.9 (no version string), only 'GET' is allowed */
	if (version == NULLCHAR && strncmp (command, "GET", 3))	{
		retval = 400;
	  	httpHeaders (HTTPhtml, "400 Bad Request", rq, 0, NULLCHAR);
		usputs (s, "<HEAD><TITLE>400 Bad Request</TITLE></HEAD>\n<BODY><H1>400 Bad Request</H1>\nYour client sent a query that this server could not understand.<P>\nReason: Invalid HTTP/0.9 method.<P></BODY>\n");
		goto quit0;
	}

	/* not complete yet, but it at least logs it and responds ;-) */
	if (!strncmp (command, "POST", 4)){
		int postsize = 0;
		int headersdone = 0;
		if (HTTPpost != 0)	{
			openHTTPlog("post");
			log (s, command);
			closeHTTPlog();
		}
		while (alarm (Httptdisc * 1000L),recvline (s, rq->buf, sizeof(rq->buf)) != -1) {
			alarm (0L);
			if (headersdone)
				postsize -= strlen(rq->buf);
			rip (rq->buf);
			if (HTTPpost != 0)	{
				openHTTPlog("post");
				log (s, rq->buf);
				closeHTTPlog();
			}
			if (!strnicmp (rq->buf, HTTPHdrs[1], strlen(HTTPHdrs[1])))	{
				postsize = atoi(&rq->buf[strlen(HTTPHdrs[1])]);
			} else if (!*rq->buf && !headersdone)	{
				headersdone = 1;
				if (!postsize)	{
					retval = 400;
					httpHeaders (HTTPhtml, "400 Bad Request", rq, 0, NULLCHAR);
					usputs (s, "<HEAD><TITLE>400 Bad Request</TITLE></HEAD>\n<BODY><H1>400 Bad Request</H1>\nYour client sent an improperly formatted POST - no size given.</BODY>\n");
					goto quit0;
				}
			}
			if (headersdone && postsize <= 0)
				break;
		}
		alarm (0L);
		retval = 202;
		httpHeaders (HTTPhtml, "202 Accepted", rq, 0, NULLCHAR);
		usputs (s, "<HEAD><TITLE>202 Post Accepted</TITLE></HEAD>\n<BODY><H1>202 Post Accepted</H1>\nThe TNOS server has accepted and logged your posting.</BODY>\n");
		goto quit0;
	}

	if (!strncmp (command, "HEAD", 4))
		head = 1;

	if (version != NULLCHAR && !strnicmp (version, "HTTP/", 5)) {
		/* we must wait to receive header lines, until a blank line */
		while (alarm (Httptdisc * 1000L), recvline (s, rq->buf, sizeof(rq->buf)) != -1) {
			alarm (0L);
			rip (rq->buf);
			if (!*rq->buf)
				break;
			if (!strnicmp (rq->buf, HTTPHdrs[HDR_LENGTH], strlen(HTTPHdrs[HDR_LENGTH]))) {
				rq->qsize = atoi (&rq->buf[strlen(HTTPHdrs[HDR_LENGTH]) + 1]);
				if (HTTPmisc == 0)
					continue;
				openHTTPlog("mischttp");
			} else if (!strnicmp (rq->buf, HTTPHdrs[HDR_SINCE], strlen(HTTPHdrs[HDR_SINCE]))) {
				rq->newcheck = strdup (&rq->buf[strlen(HTTPHdrs[HDR_SINCE]) + 1]);
				if (HTTPmisc == 0)
					continue;
				openHTTPlog("mischttp");
			} else if (!strnicmp (rq->buf, HTTPHdrs[HDR_FROM], strlen(HTTPHdrs[HDR_FROM]))) {
				rq->from = strdup (&rq->buf[strlen(HTTPHdrs[HDR_FROM]) + 1]);
				if (HTTPmisc == 0)
					continue;
				openHTTPlog("mischttp");
			} else if (!strnicmp (rq->buf, HTTPHdrs[HDR_AUTHORIZE], strlen(HTTPHdrs[HDR_AUTHORIZE]))) {
				rq->passwd = strdup (&rq->buf[strlen(HTTPHdrs[HDR_AUTHORIZE]) + 7]);
				if (HTTPmisc == 0)
					continue;
				openHTTPlog("mischttp");
			} else if (!strnicmp (rq->buf, HTTPHdrs[HDR_REFERER], strlen(HTTPHdrs[HDR_REFERER]))) {
				if (HTTPreferer == 0)
					continue;
				openHTTPlog("referer");
				strcpy (rq->buf, &rq->buf[9]);
				strcat (rq->buf, " -> /");
				strcat (rq->buf, rq->url);
				rq->referer = strdup(&rq->buf[strlen(HTTPHdrs[HDR_REFERER]) + 1]);
			} else if (!strnicmp (rq->buf, HTTPHdrs[HDR_AGENT], strlen(HTTPHdrs[HDR_AGENT])))	{
				if (HTTPagent == 0)
					continue;
				openHTTPlog("agent");
				strcpy (rq->buf, &rq->buf[12]);
				strcat (rq->buf, " -> /");
				strcat (rq->buf, rq->url);
				rq->agent = strdup(&rq->buf[strlen(HTTPHdrs[HDR_AGENT]) + 1]);
			} else	{
				if (HTTPmisc == 0)
					continue;
				openHTTPlog("mischttp");
			}
			log (s, rq->buf);
			closeHTTPlog();
		}
		alarm (0L);
	}
	
	if (!head && strncmp (command, "GET", 3))	{
		/* not GET, HEAD, or POST, so we don't support it.... */
		retval = 400;
	  	httpHeaders (HTTPhtml, "400 Bad Request", rq, 0, NULLCHAR);
		usputs (s, "<HEAD><TITLE>400 Bad Request</TITLE></HEAD>\n<BODY><H1>400 Bad Request</H1>\nYour client sent a query that this server could not understand.<P>\nReason: Invalid or unsupported method.<P></BODY>\n");
		goto quit0;		
	}

#ifdef HTTPPBBS
	if (!strncmp (newcommand, "bbs", 3))	{
		/* if no authentication, it fails */
		if (rq->passwd == NULLCHAR)	{
			goto failauthorize;
		} else {
			/* check the authentication, here */
			char *tmp64, *cp2, pathbuf[MBXLINE], *path = pathbuf;
			if ((tmp64 = base64ToStr(rq->passwd)) == NULLCHAR)
				goto failauthorize;
#if 0
			log (-1, "HTTP Authentication Test: '%s' - '%s'", rq->passwd, (tmp64) ? tmp64 : "unknown");
#endif
			if ((cp2 = strchr (tmp64, ':')) == NULLCHAR)
				goto failauthorize0;

			*cp2++ = 0;
			if((rq->privs = userlogin(tmp64,cp2,&path,MBXLINE,&rq->anony)) == -1 || (rq->privs & NO_HTTP_IP))	{
				/* invalid authentication */
failauthorize0:			free (tmp64);
failauthorize:			retval = 401;
				httpHeaders (HTTPhtml, "401 Unauthorized", rq, 0, "the TNOS PBBS");
				tputs ("<HEAD><TITLE>Authorization Failed</TITLE></HEAD>\n<BODY><H1>Authorization Failed</H1>\nThe TNOS server requires proper authorization for this URL. Either your browser does not perform authorization, or your authorization has failed.</BODY>\n");
				goto quit0;
			}
			rq->username = strdup (tmp64);
			rq->bbspath = strdup (pathbuf);
			free (tmp64);
		}
		/* passed authentication - now check for special URLs */
		
		/* if URL is 'bbs' or 'bbs/', then change it to 'bbs/${username}.html' */
		if (strlen(newcommand) == 3 || !strcmp (newcommand, "bbs/"))	{
			retval = 302;
			sprintf (rq->buf, "bbs/%s.html", rq->username);
			free (rq->url);
			rq->url = strdup (rq->buf);
		  	httpHeaders (HTTPhtml, "302 Found", rq, 0, NULLCHAR);
		  	if (!head)
				tprintf ("<TITLE>Document moved</TITLE><H1>Document moved</H1>Please,"
				   "\n<A HREF=\"http://%s/%s\">click here</A>\n",
				   Hostname,rq->url);
			goto quit0;
		}

		/* if URL is 'bbs/area', then change it to 'bbs/area/${username}.html' */
		if (!strcmp (newcommand, "bbs/area"))	{
			retval = 302;
			sprintf (rq->buf, "bbs/area/%s.html", rq->username);
			free (rq->url);
			rq->url = strdup (rq->buf);
		  	httpHeaders (HTTPhtml, "302 Found", rq, 0, NULLCHAR);
		  	if (!head)
				tprintf ("<TITLE>Document moved</TITLE><H1>Document moved</H1>Please,"
				   "\n<A HREF=\"http://%s/%s\">click here</A>\n",
				   Hostname, rq->url);
			goto quit0;
		}

		/* if URL is 'bbs/message/areaname', then change it to 'bbs/area/areaname.html' */
		if (!strncmp (newcommand, "bbs/message/", 12) && !strstr (newcommand, ".html"))	{
			cp = &newcommand[12];
			sprintf (rq->buf, "bbs/area/%s.html", cp);
			strcpy (rq->buf, rq->buf);
			free (rq->url);
			rq->url = strdup (rq->buf);
			retval = 302;
		  	httpHeaders (HTTPhtml, "302 Found", rq, 0, NULLCHAR);
		  	if (!head)
				tprintf ("<TITLE>Document moved</TITLE><H1>Document moved</H1>Please,"
				   "\n<A HREF=\"http://%s/%s\">click here</A>\n",
				   Hostname, rq->url);
			goto quit0;
		}

		/* if URL is 'bbs/$(username}.html', then build it */
		sprintf (rq->buf, "bbs/%s.html", rq->username);
		if (!strcmp (rq->buf, newcommand))	{
			/* special URL for user's login */
			rq->isvolatile = 1;
		  	httpHeaders (HTTPhtml, "200 Document follows", rq, 0, NULLCHAR);
		  	addUserFile (rq);
			goto quit0;
		}

		/* if URL is 'bbs/area/${areaname}.html', then build an area feature page */
		if (!strncmp (newcommand, "bbs/area/", 9) && !strchr(&newcommand[9], '/') && strstr (newcommand, ".html"))	{
			char *cp2;
			/* special URL for area listing */
		  	strcpy (rq->buf, &newcommand[9]);
		  	cp2 = strstr (rq->buf, ".html");
		  	*cp2 = 0;
		  	rq->areaname = strdup (rq->buf);

		  	if (!checkareaperms (rq))
		  		goto failauthorize;
		  	
			rq->isvolatile = 1;
		  	httpHeaders (HTTPhtml, "200 Document follows", rq, 0, NULLCHAR);
			sprintf (rq->buf, "%s/bbs/areaopt.html", rq->dirname);
		  	addMsgFile (rq->buf, rq);
			goto quit0;
		}

		/* if URL is 'bbs/area/areaname/xxx.html', then build an area list, starting w/msg#xxx */
		if (!strncmp (newcommand, "bbs/area/", 9))	{
			char *cp2, *cp3;
			/* special URL for area listing */
		  	strcpy (rq->buf, &newcommand[9]);
		  	if ((cp2 = strchr (rq->buf, '/')) == NULLCHAR)
		  		goto failauthorize;
		  	*cp2++ = 0;
		  	rq->areaname = strdup (rq->buf);

		  	if (!checkareaperms (rq))
		  		goto failauthorize;
		  	
		  	if ((cp3 = strchr (cp2, '.')) == NULLCHAR)
		  		goto failauthorize;
		  	*cp3 = 0;

		  	rq->msgnum = atoi(cp2);
		  	if (!rq->msgnum && !strcmp (cp2, "new"))
		  		rq->msgnum = -1;
		  	
			doPBBScmd_init (rq);
			rq->msgcount = rq->m->nmsgs;

			rq->isvolatile = 1;
		  	httpHeaders (HTTPhtml, "200 Document follows", rq, 0, NULLCHAR);

			strcpy (rq->m->line, "L");
			if (rq->msgnum != -1)
				sprintf (&rq->m->line[1], " %d %d", rq->msgnum, rq->msgnum + 99);

			sprintf (rq->buf, "%s/bbs/arealist.html", rq->dirname);
		  	addMsgFile (rq->buf, rq);
			goto quit0;
		}

		/* if URL is 'bbs/message/areaname/xxx.html', then build an message page */
		if (!strncmp (newcommand, "bbs/message/", 12))	{
			char *cp2, *cp3;
			/* special URL for reading messages */
		  	strcpy (rq->buf, &newcommand[12]);
		  	if ((cp2 = strchr (rq->buf, '/')) == NULLCHAR)
		  		goto failauthorize;
		  	*cp2++ = 0;
		  	rq->areaname = strdup (rq->buf);

		  	if (!checkareaperms (rq))
		  		goto failauthorize;
		  	
		  	if ((cp3 = strchr (cp2, '.')) == NULLCHAR)
		  		goto failauthorize;
		  	*cp3 = 0;
			rq->msgnum = atoi (cp2);

			doPBBScmd_init (rq);
			rq->msgcount = rq->m->nmsgs;

			rq->isvolatile = 1;
		  	httpHeaders (HTTPhtml, "200 Document follows", rq, 0, NULLCHAR);

			sprintf (rq->m->line, "%d", rq->msgnum);
			sprintf (rq->buf, "%s/bbs/msglist.html", rq->dirname);
		  	addMsgFile (rq->buf, rq);
			goto quit0;
		}

		/* if URL is 'bbs/cmd/xxxx', then build an text page from the BBS output of 'xxxx' */
		if (HTTPexecbbs && !strncmp (newcommand, "bbs/cmd/", 8))	{
			char *cp2, *cp3;

			cp2 = &newcommand[8];
			if (!strncmp (cp2, "area?", 5))	{
				cp3 = strchr (&cp2[5], '/');
				if (cp3 != NULLCHAR)	{
					cp2 += 5;
					*cp3++ = 0;
					rq->areaname = strdup (cp2);
					cp2 = cp3;
				}
			}
			while ((cp3 = strchr (cp2, '?')) != NULLCHAR)
		  		*cp3 = ' ';

			rq->isvolatile = 1;
		  	httpHeaders (HTTPplain, "200 Document follows", rq, 0, NULLCHAR);

			doPBBScmd (rq, cp2);
			goto quit0;
		}
		/* otherwise, it must be a regular html file in the 'bbs' directory */
	}
#endif

#ifdef MSDOS
	if (length == 0 || !strcmp(&newcommand[1],":/"))	{
		sprintf(rq->buf,"%c:%s/root.html", (newcommand[1] == ':') ? newcommand[0] : DR, rq->dirname);
	} else if (newcommand[1] == ':') {
		sprintf (rq->buf, "%c:%s%s", newcommand[0], rq->dirname, &newcommand[2]);
	} else {
		sprintf (rq->buf, "%c:%s/%s", DR, rq->dirname, newcommand);
	}
#else
	if (length == 0)	{
		sprintf(rq->buf,"%s/root.html", rq->dirname);
	} else {
		sprintf(rq->buf,"%s/%s", rq->dirname, newcommand);
	}
#endif

	/* delete any .. to avoid */
	/* anyone telnetting to http and cd'ing to a directory */
	/* other than within Httpdir */

	for ( i=0; i <= (int) strlen(rq->buf); i++ ) {
		if (rq->buf[i] == '.' && rq->buf[i+1] == '.') {
			rq->buf[i] = ' ';
			rq->buf[i+1] = ' ';
		}
	}

	if (stat (rq->buf, &sb) != -1)
		isdir = sb.st_mode & S_IFDIR;
	if (rq->buf[strlen(rq->buf)-1] == '/') {
		strcat (rq->buf, "welcome.html");
		if ((fp = fopen (rq->buf, READ_BINARY)) != NULLFILE)	{
			retval = 200;
		  	httpHeaders (HTTPhtml, "200 Document follows", rq, 0, NULLCHAR);
		  	if (!head)
				err = mkwelcome (rq->dirname, rq->buf);
		}
	} else if (isdir) {
		retval = 302;
		strcpy (rq->buf, newcommand);
		strcat (rq->buf, "/");
		free (rq->url);
		rq->url = strdup (rq->buf);
	  	httpHeaders (HTTPhtml, "302 Found", rq, 0, NULLCHAR);
	  	if (!head)
			tprintf ("<TITLE>Document moved</TITLE><H1>Document moved</H1>Please,"
			   "\n<A HREF=\"http://%s/%s\">click here</A>\n",
			   Hostname, rq->url);
		goto quit0;
	} else
		fp = fopen (rq->buf, READ_BINARY);

	if (fp != NULLFILE)	{
		int thetype = -1;

		if ((cp = strrchr (rq->buf, '.')) != NULLCHAR)	{
			cp++;
			for (thetype = HTTPhtml; HTTPtypes[thetype].ext != NULLCHAR; thetype++) {
				if (!strnicmp (HTTPtypes[thetype].ext, cp, strlen(HTTPtypes[thetype].ext)))
					break;
			}
		}
		if (thetype == -1 || HTTPtypes[thetype].type1 == -1)
			/* either had no extension, or extension not found in table */
			thetype = (isbinary(fp)) ? HTTPbinary : HTTPplain;

		retval = 200;
		(void) fstat (fileno(fp), &sb);
		if (!head && rq->newcheck && !isnewer (sb.st_mtime, rq->newcheck))	{
			head = 1;
			retval = 304;
		}

		if (!strnicmp (newcommand, "bbs/user/", 9))
			rq->isvolatile = 1;
	  	httpHeaders (thetype, (retval == 304) ? "304 Not Modified" : "200 Document follows", rq, 1, NULLCHAR);
	  	httpFileinfo (fp, rq, thetype, version);
		if (!head)	{
			if (thetype != HTTPhtml)	/* if not html, just send */
				(void) sendfile (fp, s, IMAGE_TYPE, 0);
			else				/* otherwise, scan for server-side includes */
				sendhtml (fp, rq);
		}
		fclose (fp);
	} else if (err) {
		retval = 404;
		httpHeaders (HTTPhtml, "404 Not Found", rq, 0, NULLCHAR);
		if (!head)
			tprintf ("<TITLE>Error</TITLE><H1>404 Not Found</H1>The file \"/%s\" does not exist on this server.\n", newcommand);
	}
quit0:
	free (tmp);
	openHTTPlog ("http");
	log (s, "port #%u: \"%s\" %d", port, command, retval);
	closeHTTPlog();
	(void) countusage ("tcount.dat", rq->dirname, 0, 1);
 
quit:
        usflush(Curproc->output);
	close_s (Curproc->output);

	free (rq->url);
	free (rq->method);
	free (rq->version);
	free (rq->newcheck);
	free (rq->from);
	free (rq->referer);
	free (rq->agent);
	free (rq->passwd);
	free (rq->usetime);
	free (rq->useerror);
	
#ifdef HTTPPBBS
	free (rq->areaname);
	free (rq->username);
	free (rq->bbspath);
#endif
	free (rq);

	HttpUsers--;
	psignal (&HttpUsers, 1);
}

char *
gmt_ptime (thetime)
time_t *thetime;
{
	return ht_time(*thetime,HTTP_TIME_FORMAT, 1);
}

static char *
ht_time (t, fmt, gmt)
time_t t;
const char *fmt;
int gmt;
{
static char ts[LINELEN];
struct tm *tms;

	tms = (gmt ? gmtime(&t) : localtime(&t));

	/* check return code? */
	strftime (ts, LINELEN, fmt, tms);
	return ts;
}


static void
httpFileinfo (fp, rq, thetype, vers)
FILE *fp;
struct reqInfo *rq;
int thetype;
char *vers;
{
struct stat sb;

	if (vers == NULLCHAR)
		return;		/* it is 0.9, no headers sent */
	if (!rq->isvolatile)
		if (fstat (fileno(fp), &sb) != -1)
			tprintf ("%s %s", HTTPHdrs[HDR_MODIFIED], gmt_ptime(&sb.st_mtime));

	if (thetype != HTTPhtml)
		tprintf ("%s %lu\n", HTTPHdrs[HDR_LENGTH], filelength(fileno(fp)));
	tprintf ("\n");
}

static void
httpHeaders (type, str, rq, more, realm)
int type, more;
struct reqInfo *rq;
const char *str, *realm;
{
long thetime;

	if (rq->version == NULLCHAR)
		return;		/* it is 0.9, no headers sent */
	(void) time (&thetime);
	
	tprintf ("HTTP/1.0 %s\n%s %s", str, HTTPHdrs[HDR_DATE], gmt_ptime(&thetime));
	if (rq->isvolatile)
		tprintf ("%s %s", HTTPHdrs[HDR_EXPIRES], gmt_ptime(&thetime));
	if (realm)
		tprintf ("%s basic realm=\"%s\"\n", HTTPHdrs[HDR_AUTHENTICATE], realm);
	tprintf ("%s 1.0\n%s TNOS/"VERSION" "OS"\n", HTTPHdrs[HDR_MIME], HTTPHdrs[HDR_SERVER]);
	tprintf ("%s http://%s/%s\n", HTTPHdrs[HDR_LOCATION], Hostname, rq->url);
	tprintf ("%s %s%s\n", HTTPHdrs[HDR_TYPE], CTypes1[HTTPtypes[type].type1], HTTPtypes[type].type2);

	/* must have this line, when no more headers needed */
	if (more == 0)
		tputs ("\n");
}


static int
mkwelcome (Hdir, file)
const char *Hdir;
char *file;
{
char *dirstring,*p,*cp;
const char *ccp, *bitmap;
struct ffblk ffblk;
int done,pre = 0, k;
char *updir = NULLCHAR;

	if ((dirstring = (char *)mallocw (strlen(file) + 16)) == NULL)
		return 0;

	strcpy (dirstring, file);
	cp = strrchr (dirstring, '/'); 
	*++cp = '\0';

	if (strlen(Hdir) != strlen(dirstring))	{
		updir = strdup(&dirstring[strlen(Hdir) + 1]);
		p = strrchr (updir, '/');
		*p = 0;
		p = strrchr (updir, '/');
		if (p == NULLCHAR)
			updir[0] = 0;
		else
			*++p = 0;
	}
	
	strcat (dirstring, "wwwpre.cat");
	pre = addFile (dirstring, Curproc->output);

	*cp = '\0';
	strcat (dirstring, WILDCARD);
	done = findfirst (dirstring, &ffblk, FA_DIREC);

	(void) sockmode (Curproc->output, SOCK_ASCII);
	(void) setflush (Curproc->output, -1); /* we will do our own flushing */
	if (!done) {
		*cp = '\0';
#ifdef MSDOS
		p = dirstring + strlen(Hdir) + 2;
#else
		p = dirstring + strlen(Hdir) + 1;
#endif
		undosify (p);
		if (!pre)
#ifdef MSDOS
			tprintf ("<HEAD><TITLE>Index of /%c:%s</TITLE></HEAD></BODY><H1>Index of /%c:%s</H1>\n", file[0], p, file[0], p);
#else
			tprintf ("<HEAD><TITLE>Index of /%s</TITLE></HEAD></BODY><H1>Index of /%s</H1>\n", p, p);
#endif
		tputs ("<PRE>\n");
	} else {
		free (dirstring);
		return -1;
	}

	tputs ("<IMG SRC=\"/icons/blank.xbm\" ALT=\"    \"> Name                   Last modified     Size\n<HR>\n");
	if (updir != NULLCHAR)	{
#ifdef MSDOS
		tprintf ("<IMG SRC=\"/icons/menu.xbm\" ALT=\"[DIR]\"> <A HREF=\"/%c:%s\">Parent Directory</a>\n", file[0], updir);
#else
		tprintf ("<IMG SRC=\"/icons/menu.xbm\" ALT=\"[DIR]\"> <A HREF=\"/%s\">Parent Directory</a>\n", updir);
#endif
		free (updir);
	}
	while (!done) {
		(void) strlwr (ffblk.ff_name);
		if (ffblk.ff_name[0] != '.') {
			if (ffblk.ff_attrib == FA_DIREC)	{
				ccp = "[DIR]";
				bitmap = "menu";
			} else {
				ccp = "[FILE]";
				bitmap = "text";
				k = strlen(ffblk.ff_name) - 4;
				if (k > 0 && (!strnicmp (&ffblk.ff_name[k], ".gif", 4) || !strnicmp (&ffblk.ff_name[k], ".jpg", 4) || !strnicmp (&ffblk.ff_name[k], ".xbm", 4)))
					bitmap = "image";
			}
			tprintf (entry, bitmap, ccp);
			tprintf (entry1, 
#ifdef MSDOS
				file[0],
#endif
				p,ffblk.ff_name,ffblk.ff_name);
			for (k = strlen (ffblk.ff_name); k < 22; k++)
				tputc (' ');
			tprintf (entry2,
#ifdef MSDOS
				(ffblk.ff_fdate) & 0x1f,	/* day */
				Months[((ffblk.ff_fdate >> 5) & 0xf)-1], /* month */
				(ffblk.ff_fdate >> 9) + 80,	/* year */
				(ffblk.ff_ftime >> 11) & 0x1f,	/* hour */
				(ffblk.ff_ftime >> 5) & 0x3f,	/* minutes */
#else
				ffblk.ff_ftime.tm_mday,		/* day */
				Months[ffblk.ff_ftime.tm_mon],	/* month */
				ffblk.ff_ftime.tm_year,		/* year */
				ffblk.ff_ftime.tm_hour,		/* hour */
				ffblk.ff_ftime.tm_min,		/* minute */
#endif
				(int)((ffblk.ff_fsize + 1023L) / 1024L));
		}
		pwait (NULL);
		done = findnext (&ffblk);
	}
	tputs ("</PRE>\n");
	usflush (Curproc->output);

	strcat (dirstring, "wwwpost.cat");
	if (!addFile (dirstring, Curproc->output))
		tputs ("</BODY>\n");
	free (dirstring);
	return 0;
}


extern FILE *Logfp;

static void
openHTTPlog (name)
const char *name;
{
char filename[256];

	HTTPsavedfp = Logfp;
	sprintf (filename, "%s/%s.log", LOGdir, name);
	Logfp = fopen (filename, APPEND_TEXT);
}

static void
closeHTTPlog (void)
{
	if (Logfp != NULLFILE)
		fclose (Logfp);
	Logfp = HTTPsavedfp;
}

static int
getmonth (cp)
char *cp;
{
int k;

	for (k = 0; k < 12; k++)	{
		if (!strnicmp (cp, Months[k], 3))	{
			return k;
		}
	}
	return 0;
}


static long
countusage (file, basedir, display, increase)
const char *file, *basedir;
int display, increase;
{
char name[256], buf[32];
FILE *fp;
long count = 0;

#ifdef MSDOS
	sprintf (name,"%c:%s/counts/%s", DR, basedir, file);
#else
	sprintf (name,"%s/counts/%s", basedir, file);
#endif
	if ((fp = fopen (name, READ_TEXT)) != NULLFILE) {
		(void) fgets (buf, sizeof(buf), fp);
		fclose (fp);
		count = atol (buf);
	}
	if (increase)	{
		count++;
		if ((fp = fopen (name, WRITE_TEXT)) != NULLFILE) {
			fprintf (fp, "%ld", count);
			fclose (fp);
		}
	}
	if (display)
		tprintf ("%ld", count);
	return (count);
}


static int
isnewer (thetime, tmstr)
time_t thetime;
char *tmstr;
{
time_t new;
struct tm t, *ltm;
char *cp, *cp2;
int format = 0;
time_t offset;

	/* the string should be in one of these three formats:
	 * (a) RFC 822 (preferred) - Sun, 06 Nov 1994 08:49:37 GMT
	 * (b) RFC 850		   - Sunday, 06-Nov-94 08:49:37 GMT
	 * (c) ANSI C's asctime    - Sun Nov  6 08:49:37 1994
	 */
	/* skip past the name of the weekday */
	cp = skipwhite (tmstr);
	cp = skipnonwhite (cp);
	cp = skipwhite (cp);
	if (isdigit (*cp))	{	/* format a or b */
		cp[2] = 0;
		t.tm_mday = atoi (cp);
		cp += 3;
		t.tm_mon = getmonth (cp);
		cp += 4;
		cp2 = skipnonwhite (cp);
		*cp2 = 0;
		t.tm_year = atoi (cp);
		cp = ++cp2;
	} else {			/* format c */
		t.tm_mon = getmonth (cp);
		cp += 4;
		t.tm_mday = atoi (cp);
		cp += 3;
		format = 1;
	}

	/* time parsing is common between the three formats */
	sscanf (cp, "%02d:%02d:%02d", &t.tm_hour, &t.tm_min, &t.tm_sec);
	/* then parse the year, if format c */
	if (format) {
		cp += 9;
		t.tm_year = atoi (cp);
	}
	if (t.tm_year > 1900)
		t.tm_year -= 1900;
#if !defined(sun) && !defined(__bsdi__)
	t.tm_isdst = -1;
#endif
	new = mktime (&t);
	ltm = localtime(&new);		/* to get timezone info */
#if defined(sun) || defined(__bsdi__)
	offset = (ltm->tm_gmtoff * -1);
#else
	offset = ((timezone - (ltm->tm_isdst * 3600L)) * -1);
#endif
	new += offset;
	return (new != thetime);
}

static void
send_size (sizefmt, size)
int sizefmt;
long size;
{
	if (sizefmt == SIZEFMT_BYTES)	{
		tprintf ("%ld", size);
		return;
	}
	if (size == -1)
		tputs ("    -");
	else {
		if (!size)
			tputs ("   0K");
		else if (size < 1024)
			tputs ("   1K");
		else if (size < 1048576)
			tprintf ("%4ldK", size / 1024L);
		else
			tprintf ("%4ldM", size / 1048576L);
	}
}


static void
sendhtml (fp, rq)
FILE *fp;
struct reqInfo *rq;
{
time_t t;
struct stat ft;
char *cp1, *cp2, c;
char *remainder, *cmdname, newname[256];
int val;
FILE *tfp;
char *bufptr;
char buf[PLINELEN + 16];

	(void) sockmode (Curproc->output, SOCK_BINARY);
    	while ((fgets (buf, PLINELEN + 16, fp)) != NULLCHAR) {
		pwait(NULL);
		bufptr = buf;
		while (bufptr  && (cp1 = strstr (bufptr, "<!--#")) != NULLCHAR) {
			if (cp1 != bufptr)	{
				*cp1 = 0;
				tputs (bufptr);
			}
			cp1 += 5;
			remainder = strstr (cp1, "-->");
			if (remainder != NULLCHAR) {
				*remainder = 0;
				remainder = &remainder[3];
			}
			cmdname = skipwhite (cp1);
			cp1 = skipnonwhite (cmdname);
			cp1 = skipwhite (cp1);

			if (!strnicmp (cmdname, "echo", 4)) {
				if (!strnicmp (cp1, "var=\"", 5)) {
					(void) time (&t);
					cp1 += 5;
					if ((cp2 = strrchr (cp1, '\"')) != NULLCHAR)
						*cp2 = 0;
					if (!stricmp (cp1, "DATE_LOCAL"))
						tputs (ht_time (t, rq->usetime, 0));
					else if (!stricmp (cp1, "DATE_GMT"))
						tputs (ht_time (t, HTTP_TIME_FORMAT, 1));
					else if (!stricmp (cp1, "HOSTNAME"))
						tputs ((Hostname) ? Hostname : "unknownhost");
					else if (!stricmp (cp1, "DOCUMENT_URI"))
						tprintf ("/%s", rq->url);
					else if (!stricmp (cp1, "VERSION"))
						tputs (Version);
#ifdef HTTPPBBS
					else if (!stricmp (cp1, "USERNAME"))
						tputs ((rq->username) ? rq->username : "unknownUser");
					else if (!stricmp (cp1, "AREANAME"))
						tputs ((rq->areaname) ? rq->areaname : "unknownArea");
					else if (!stricmp (cp1, "MESSAGENUM"))
						tprintf ("%d", rq->msgnum);
					else if (!stricmp (cp1, "PREV_MSGNUM"))
						tprintf ("%d", rq->msgnum - 1);
					else if (!stricmp (cp1, "NEXT_MSGNUM"))
						tprintf ("%d", rq->msgnum + 1);
					else if (!stricmp (cp1, "PERMISSIONS"))
						doPBBScmd (rq, "security");
					else if (!stricmp (cp1, "PREV_LIST_START"))
						tprintf ("%d", (rq->msgnum > 100) ? rq->msgnum - 100 : 1);
					else if (!stricmp (cp1, "NEXT_LIST_START"))
						tprintf ("%d", rq->msgnum + 100);
					else if (!stricmp (cp1, "PREV_LIST_BUTTON"))	{
						if (rq->msgnum > 1)	{
							sprintf (rq->buf, "%s/bbs/plbutton.cat", rq->dirname);
							if ((tfp = fopen (rq->buf, "r")) != NULLFILE) {
								sendhtml (tfp, rq);
								fclose (tfp);
							}
						}
					} else if (!stricmp (cp1, "PREV_MSG_BUTTON"))	{
						if (rq->msgnum > 1)	{
							sprintf (rq->buf, "%s/bbs/pmbutton.cat", rq->dirname);
							if ((tfp = fopen (rq->buf, "r")) != NULLFILE) {
								sendhtml (tfp, rq);
								fclose (tfp);
							}
						}
					} else if (!stricmp (cp1, "NEXT_MSG_BUTTON"))	{
						if (rq->msgcount > rq->msgnum)	{
							sprintf (rq->buf, "%s/bbs/nmbutton.cat", rq->dirname);
							if ((tfp = fopen (rq->buf, "r")) != NULLFILE) {
								sendhtml (tfp, rq);
								fclose (tfp);
							}
						}
					} else if (!stricmp (cp1, "NEXT_LIST_BUTTON"))	{
						if (rq->msgcount > rq->msgnum + 100)	{
							sprintf (rq->buf, "%s/bbs/nlbutton.cat", rq->dirname);
							if ((tfp = fopen (rq->buf, "r")) != NULLFILE) {
								sendhtml (tfp, rq);
								fclose (tfp);
							}
						}
					} else if (!stricmp (cp1, "AREA_LIST") || !stricmp (cp1, "AREA_MSG"))
						doPBBScmd_term (rq);
#endif
#ifdef ALLSERV
					else if (!stricmp (cp1, "QUOTE")) {
						cp2 = getquote ();
						tputs ((cp2) ? cp2 : "[quote not found]");
					}
#endif
					else if (!stricmp (cp1, "DOCUMENT_NAME"))	{
						if (rq->url)	{
							cp2 = strrchr (rq->url, '/');
							if (cp2)
								cp2++;
							else
								cp2 = rq->url;
							tputs (cp2);
						}
					} else if (!stricmp (cp1, "LAST_MODIFIED")) {
						(void) fstat (fileno(fp), &ft);
						tputs (ht_time (ft.st_mtime, rq->usetime, 0));
					} else if (!stricmp (cp1, "TOTAL_HIT_COUNT") || !stricmp (cp1, "TOTAL_HITS"))
						(void) countusage ("tcount.dat", rq->dirname, 1, 0);
					else if (!stricmp (cp1,"REQ_FROM")) {
						if (rq->from)
							tputs(rq->from);
					} else if (!stricmp (cp1,"REQ_REFERER")) {
						if (rq->referer)
							tputs (rq->referer);
					} else if (!stricmp (cp1,"REQ_AGENT")) {
						if (rq->agent)
							tputs (rq->agent);
					} else
						tputs (rq->useerror);
				} else if (!strnicmp (cp1, "count=\"", 7) || !strnicmp (&cp1[1], "count=\"", 7))	{
					c = tolower(*cp1);
					switch (c) {
						case 'c':	/* count */
						case 'i':	/* icount */
						case 's':	/* scount */
								cp1 = strchr (cp1, '\"');
								cp2 = strrchr (++cp1, '\"');
								if (cp2)
									*cp2 = 0;
								(void) countusage (cp1, rq->dirname, (c != 's'), (c != 'c'));
								break;
						default:	tputs (rq->useerror);

					}
				} else
					tputs (rq->useerror);
			} else if (!strnicmp (cmdname, "config", 6)) {
				if (!strnicmp (cp1, "timefmt=\"", 9)) {
					cp1 += 9;
					if ((cp2 = strrchr (cp1, '\"')) != NULLCHAR)
						*cp2 = 0;
					free (rq->usetime);
					rq->usetime = strdup (cp1);
				} else if (!strnicmp (cp1, "sizefmt=\"", 9)) {
					c = tolower(cp1[9]);
					/* either 'abbrev' or 'bytes' */
					rq->sizefmt = (c == 'a') ? SIZEFMT_KMG : SIZEFMT_BYTES;
				} else if (!strnicmp (cp1, "errmsg=\"", 8)) {
					cp1 += 8;
					if ((cp2 = strrchr (cp1, '\"')) != NULLCHAR)
						*cp2 = 0;
					free (rq->useerror);
					rq->useerror = strdup (cp1);					
				} else
					tputs (rq->useerror);
			} else if (!strnicmp (cmdname, "include", 7)) {
				file_or_virtual (cp1, newname, rq);
				if (*newname && (tfp = fopen (newname, "r")) != NULLFILE) {
					sendhtml (tfp, rq);
					fclose (tfp);
				} else
					tputs (rq->useerror);
			} else if (!strnicmp (cmdname, "fsize", 5)) {
				file_or_virtual (cp1, newname, rq);
				if (*newname)	{
					(void) stat (newname, &ft);
					send_size (rq->sizefmt, (long) ft.st_size);
				} else
					tputs (rq->useerror);
                        } else if (!strnicmp (cmdname, "flastmod", 8)) {
				file_or_virtual (cp1, newname, rq);
				if (*newname)	{
					(void) stat (newname, &ft);
					tputs (ht_time (ft.st_mtime, rq->usetime, 0));
				} else
					tputs (rq->useerror);
#if 0
			} else if (HTTPscript && !strnicmp (cmdname, "tscript", 7)) {
				/* to be added later */
				tputs (rq->useerror);
#endif
			} else if (!strnicmp (cmdname, "exec", 4))	{
				/* exec cgi no supported, at least not yet */
#ifdef HTTPPBBS
				if (HTTPexecbbs && !strnicmp (cp1, "bbs=\"", 5)) {
					if ((cp2 = strrchr (cp1, '\"')) != NULLCHAR)
						*cp2 = 0;
					doPBBScmd (rq, &cp1[5]);
				} else
#endif
#ifdef UNIX
				if (HTTPexeccmd && !strnicmp (cp1, "cmd=\"", 5)) {
					if ((cp2 = strrchr (cp1, '\"')) != NULLCHAR)
						*cp2 = 0;
					if ((tfp = popen (&cp1[5], "r")) != NULLFILE) {
						while ((val = fgetc (tfp)) != EOF)
							tputc (val);
					}
					(void) pclose (tfp);
				} else
#endif
					tputs (rq->useerror);
			} else {
				tprintf ("<!--#%s-->", cmdname);
			}
			bufptr = remainder;
		}
		if (bufptr != NULLCHAR)
			tputs (bufptr);
	}
}


static void
file_or_virtual (cp, newname, rq)
char *cp, *newname;
struct reqInfo *rq;
{
char *cp2, *cp3;

	*newname = 0;
	if (!strnicmp (cp, "virtual=\"", 9))	{
		/* virtual filename, based on root http directory */
		cp2 = &cp[9];
		if (!strncmp (cp2, "../", 3) || !strncmp (cp2, "./", 2))
			cp2 = (strchr (cp2, '/') + 1);
		sprintf (newname, "%s/%s", rq->dirname, cp2);
	} else if (!strnicmp (cp, "file=\"", 6)) {
		/* absolute filename */
		cp2 = &cp[6];
		if (!strncmp (cp2, "../", 3) || !strncmp (cp2, "./", 2))
			cp2 = (strchr (cp2, '/') + 1);
		if (*cp2 == '/')
			cp2++;
		cp3 = strrchr (rq->url, '/');
		sprintf (newname, "%s%s/%s", rq->dirname, (cp3) ? cp3 : "", cp2);
	}
	if (*newname && (cp2 = strrchr (newname, '\"')) != NULLCHAR)
		*cp2 = 0;
}

#endif /* HTTP */

